//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//############################################################################
#ifndef __EZ_3D_DOUBLE_DEFINITIONS__
#define __EZ_3D_DOUBLE_DEFINITIONS__

//############################################################################
#include "ez_uail_common.hpp"
#include "ez_3D_short.hpp"

//############################################################################
class ez_3D_double
{
public:
    ez_3D_double(const double& f = 0)
               : x(f), y(f), z(f)               {}
    //------------------------------------------------------------------------
    ez_3D_double( const double& _x,
                 const double& _y,
                 const double& _z
               )
               : x(_x), y(_y), z(_z)            {}
    //------------------------------------------------------------------------
    ez_3D_double(const ez_3D_double& f)
               : x(f.x), y(f.y), z(f.z)         {}
    //------------------------------------------------------------------------
    ez_3D_double(const ez_3D_short& s)
               : x(s.x), y(s.y), z(s.z)         {}
    //------------------------------------------------------------------------
    ez_3D_double& operator = (const ez_3D_double& f)
                    {
                        x = f.x;
                        y = f.y;
                        z = f.z;
                        return *this;
                    }
    //------------------------------------------------------------------------
    ez_3D_double& operator = (const double& f)
                    {
                        x = y = z = f;
                        return *this;
                    }
    //------------------------------------------------------------------------
    bool operator == (const ez_3D_double& f) const
            {
                if(    (x != f.x)
                    || (y != f.y)
                    || (z != f.z)
                  )
                    return false;
                return true;
            }
    //------------------------------------------------------------------------
    bool operator != (const ez_3D_double& f) const
            {   
                if(    (x == f.x)
                    && (y == f.y)
                    && (z == f.z)
                  )
                    return false;
                return true;
            }
    //------------------------------------------------------------------------
    bool operator == (const double& f) const
            {
                if(    (x != f)
                    || (y != f)
                    || (z != f)
                  )
                    return false;
                return true;
            }
    //------------------------------------------------------------------------
    bool operator != (const double& f) const
            {   
                if(    (x == f)
                    && (y == f)
                    && (z == f)
                  )
                    return false;
                return true;
            }
    //------------------------------------------------------------------------
    bool operator > (const ez_3D_double& f) const // distance from the origin
            {   
                return(   (  x *   x +   y *   y +   z *   z)
                        > (f.x * f.x + f.y * f.y + f.z * f.z)
                      );
            }
    //------------------------------------------------------------------------
    bool operator >= (const ez_3D_double& f) const // distance from the origin
            {   
                return(    (  x *   x +   y *   y +   z *   z)
                        >= (f.x * f.x + f.y * f.y + f.z * f.z)
                      );
            }
    //------------------------------------------------------------------------
    bool operator < (const ez_3D_double& f) const // distance from the origin
            {   
                return(   (  x *   x +   y *   y +   z *   z)
                        < (f.x * f.x + f.y * f.y + f.z * f.z)
                      );
            }
    //------------------------------------------------------------------------
    bool operator <= (const ez_3D_double& f) const // distance from the origin
            {   
                return(    (  x *   x +   y *   y +   z *   z)
                        <= (f.x * f.x + f.y * f.y + f.z * f.z)
                      );
            }
    //------------------------------------------------------------------------
    ez_3D_double  operator -  (const ez_3D_double& f) const
                    {   
                        return ez_3D_double(x - f.x, y - f.y, z - f.z);
                    }
    //------------------------------------------------------------------------
    ez_3D_double& operator -= (const ez_3D_double& f)
                    {   
                        x -= f.x;
                        y -= f.y;
                        z -= f.z;
                        return *this;
                    }
    //------------------------------------------------------------------------
    ez_3D_double& operator -= (const double& f)
                    {   
                        x -= f;
                        y -= f;
                        z -= f;
                        return *this;
                    }
    //------------------------------------------------------------------------
    ez_3D_double  operator +  (const ez_3D_double& f) const
                    {   
                        return ez_3D_double(x + f.x, y + f.y, z + f.z);
                    }
    //------------------------------------------------------------------------
    ez_3D_double& operator += (const ez_3D_double& f)
                    {   
                        x += f.x;
                        y += f.y;
                        z += f.z;
                        return *this;
                    }
    //------------------------------------------------------------------------
    ez_3D_double& operator += (const double& f)
                    {   
                        x += f;
                        y += f;
                        z += f;
                        return *this;
                    }
    //------------------------------------------------------------------------
    ez_3D_double  operator *  (const ez_3D_double& f) const
                    {   
                        return ez_3D_double(x * f.x, y * f.y, z * f.z);
                    }
    //------------------------------------------------------------------------
    ez_3D_double  operator /  (const ez_3D_double& f) const
                    {   
                        ez_3D_double quotient;
                        if(f.x)    quotient.x = x / f.x;
                        if(f.y)    quotient.y = y / f.y;
                        if(f.z)    quotient.z = z / f.z;
                        return quotient;
                    }
    //------------------------------------------------------------------------
    ez_3D_double  operator *  (const double& f) const
                    {   
                        return ez_3D_double(x * f, y * f, z * f);
                    }
    //------------------------------------------------------------------------
    ez_3D_double  operator /  (const double& f) const
                    {   
                        if(f)
                            return ez_3D_double(x / f, y / f, z / f);
                        return ez_3D_double();
                    }
    //------------------------------------------------------------------------
    ez_3D_double  operator /  (const int& i) const
                    {   
                        if(i)
                            return ez_3D_double(x / i, y / i, z / i);
                        return ez_3D_double();
                    }
    //------------------------------------------------------------------------
    ez_3D_double& operator *= (const ez_3D_double& f)
                    {   
                        x *= f.x;
                        y *= f.y;
                        z *= f.z;
                        return *this;
                    }
    //------------------------------------------------------------------------
    ez_3D_double& operator *= (const double& f)
                    {   
                        x *= f;
                        y *= f;
                        z *= f;
                        return *this;
                    }
    //------------------------------------------------------------------------
    operator ez_3D_short() const // cast to ez_3D_short
                {
                    return ez_3D_short( (short)round(x),
                                        (short)round(y), 
                                        (short)round(z)
                                      );
                }
    //------------------------------------------------------------------------
    ez_3D_double operator - ()
                {
                    ez_3D_double negative;
                    negative.x = -x;
                    negative.y = -y;
                    negative.z = -z;
                    return negative;
                }
    //------------------------------------------------------------------------
    double operator | (const ez_3D_double& f) const // distance from each other
                {   
                    return( sqrt(   (double)(x - f.x) * (x - f.x)
                                  + (double)(y - f.y) * (y - f.y)
                                  + (double)(z - f.z) * (z - f.z)
                                )
                          );
                }
    //------------------------------------------------------------------------
    double magnitude() const // distance from origin
                {   
                    ez_3D_double origin;
                    return(origin | *this);
                }
    //------------------------------------------------------------------------
    ez_3D_double direction() const // z and y rotational angle from origin
            {   
                ez_3D_double origin,
                            basline_point,
                            copy = *this,
                            rotation;
                //............................................................
                basline_point.x = 1.0;
                copy.z          = 0.0;
                rotation.z      = (double)delta_angle(  copy | basline_point
                                                      , 1.0
                                                      , copy | origin
                                                    );
                if(rotation.y < 0) rotation.z  = two_pi - rotation.z;
                //............................................................
                copy            =  *this;
                copy.y          = 0.0;
                rotation.y      = (double)delta_angle(  copy | basline_point
                                                      , 1.0
                                                      , copy | origin
                                                    );
                if(rotation.z < 0) rotation.y  = two_pi - rotation.y;                            
                //............................................................
                return rotation;
            }
    //------------------------------------------------------------------------
    double sum()    {    return x + y + z;    }
    //------------------------------------------------------------------------
    double x, y, z;
};

//############################################################################
#endif

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
