//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//############################################################################
#ifndef __EZ_3D_REAL_DEFINITIONS__
#define __EZ_3D_REAL_DEFINITIONS__

//############################################################################
#include "ez_uail_common.hpp"
#include "ez_color.hpp"
#include "ez_3D_double.hpp"

//############################################################################
class ez_real_point : public ez_3D_double, public ez_color
{
public:
    ez_real_point( const double&  _x = 0,
                   const double&  _y = 0,
                   const double&  _z = 0,
                   const u_char&  _r = 0,
                   const u_char&  _g = 0,
                   const u_char&  _b = 0,
                   const u_char&  _k = EZ_UAIL_BLANKING_BIT,
                   const u_char&  _c = 0
                  )
                  : ez_3D_double (_x, _y, _z),
                    ez_color     (_r, _g, _b),
                    k            (_k        ),
                    c            (_c        )                   {}
    //------------------------------------------------------------------------
    ez_real_point(const ez_real_point& p)
                 : ez_3D_double (p.x, p.y, p.z),
                   ez_color     (p.r, p.g, p.b),
                   k            (p.k          ),
                   c            (p.c          )                 {}
    //------------------------------------------------------------------------
    ez_real_point( const ez_3D_double& f,
                   const ez_color&     rgb,
                   const u_char&       _k = EZ_UAIL_BLANKING_BIT,
                   const u_char&       _c = 0
                 )
                 : ez_3D_double (f  ),
                   ez_color     (rgb),
                   k            (_k ),
                   c            (_c )                           {}
    //------------------------------------------------------------------------
    ez_real_point( const ez_3D_double& f,
                   const u_char&       _c
                 )
                 : ez_3D_double (f),
                   ez_color     (0, 0, 0),
                   k            (EZ_UAIL_BLANKING_BIT),
                   c            (_c)                            {}
    //------------------------------------------------------------------------
    ez_real_point( const ez_3D_double& f)
                 : ez_3D_double (f),
                   ez_color     (0, 0, 0),
                   k            (EZ_UAIL_BLANKING_BIT),
                   c            (0)                             {}
    //------------------------------------------------------------------------
    ez_real_point( const ez_color& rgb)
                 : ez_3D_double (0, 0, 0),
                   ez_color     (rgb),
                   k            (EZ_UAIL_BLANKING_BIT),
                   c            (0)                             {}
    //------------------------------------------------------------------------
    ez_real_point& operator = (const ez_real_point& p)
                    {
                        x = p.x;
                        y = p.y;
                        z = p.z;
                        r = p.r;
                        g = p.g;
                        b = p.b;
                        k = p.k;
                        c = p.c;
                        return *this;
                    }
    //------------------------------------------------------------------------
    ez_real_point& operator = (const ez_3D_double& f) // only assign the coordinates
                    {
                        x = f.x;
                        y = f.y;
                        z = f.z;
                        return *this;
                    }
    //------------------------------------------------------------------------
    ez_real_point  operator + (const ez_3D_double& f)
                    {
                        ez_real_point sum(*this);
                        sum.x += f.x;
                        sum.y += f.y;
                        sum.z += f.z;
                        sum.k =  k;
                        sum.c =  c;
                        sum.r =  r;
                        sum.g =  g;
                        sum.b =  b;
                        return sum;
                    }
    //------------------------------------------------------------------------
    ez_real_point& operator += (const ez_3D_double& f)
                    {
                        x += f.x;
                        y += f.y;
                        z += f.z;
                        return *this;
                    }
    //------------------------------------------------------------------------
    ez_real_point  operator - (const ez_3D_double& f)
                    {
                        ez_real_point diff(*this);
                        diff.x -= f.x;
                        diff.y -= f.y;
                        diff.z -= f.z;
                        return diff;
                    }
    //------------------------------------------------------------------------
    ez_real_point& operator -= (const ez_3D_double& f)
                    {
                        x -= f.x;
                        y -= f.y;
                        z -= f.z;
                        return *this;
                    }
    //------------------------------------------------------------------------
    ez_real_point& operator *= (const ez_3D_double& f)
                    {
                        x *= f.x;
                        y *= f.y;
                        z *= f.z;
                        return *this;
                    }
    //------------------------------------------------------------------------
    ez_real_point operator * (const ez_3D_double& f)
                    {
                        ez_real_point product;
                        product.r = r;
                        product.g = g;
                        product.b = b;
                        product.x = x * f.x;
                        product.y = y * f.y;
                        product.z = z * f.z;
                        product.k = k;
                        product.c = c;
                        return product;
                    }
    //------------------------------------------------------------------------
    bool operator == (const ez_real_point& p) const
            {   
                if(    (x != p.x)
                    || (y != p.y)
                    || (z != p.z)
                    || (r != p.r)
                    || (g != p.g)
                    || (b != p.b)
                    || ((k & EZ_UAIL_BLANKING_BIT) != (p.k & EZ_UAIL_BLANKING_BIT))
                    || (c != p.c)
                  )
                    return false;
                return true;
            }
    //------------------------------------------------------------------------
    bool operator != (const ez_real_point& p) const
            {                               
                if(    (x == p.x)
                    && (y == p.y)
                    && (z == p.z)
                    && (r == p.r)
                    && (g == p.g)
                    && (b == p.b)
                    && ((k & EZ_UAIL_BLANKING_BIT) == (p.k & EZ_UAIL_BLANKING_BIT))
                    && (c == p.c)
                  )
                    return false;
                return true;
            }
    //------------------------------------------------------------------------
    ez_real_point operator - ()
                    {
                        ez_real_point negative;
                        negative.x = -x;
                        negative.y = -y;
                        negative.z = -z;
                        return negative;
                    }
    //------------------------------------------------------------------------
    ez_real_point&  blank   () {  k |=  EZ_UAIL_BLANKING_BIT; return *this; }
    ez_real_point&  unblank () {  k &= ~EZ_UAIL_BLANKING_BIT; return *this; }
    //------------------------------------------------------------------------
    bool  is_blank    () const {  return (bool)(k & EZ_UAIL_BLANKING_BIT);  }
    bool  is_lit      () const {  return !is_blank()                     ;  }
    //------------------------------------------------------------------------
    void  set_unique  ()       {  k |=  EZ_UAIL_UNIQUE_POINT;      return;  }
    void  clear_unique()       {  k &= ~EZ_UAIL_UNIQUE_POINT;      return;  }
    bool  is_unique   () const {  return (bool)(k & EZ_UAIL_UNIQUE_POINT);  }
    //------------------------------------------------------------------------
    bool  is_black () const
                    {
                        return (    is_lit()
                                 && r == 0
                                 && g == 0
                                 && b == 0
                               );
                    }
    //------------------------------------------------------------------------
    bool  is_color () const
            {
                return (    is_lit()
                         && (r + g + b)
                       );
            }
    //------------------------------------------------------------------------
    ez_3D_double position() const   {  return (ez_3D_double)(*this);  }
    //------------------------------------------------------------------------
    int color() const
                {
                    return   is_lit()
                           ? (int)(   r << 16
                                    | g <<  8
                                    | b
                                  )
                           : -1; // is blank
                }
    //------------------------------------------------------------------------
    ez_color as_ez_color() const
        {
            return (ez_color)(*this);
        }
    //------------------------------------------------------------------------
    u_char  k, c;
};

//############################################################################
#endif

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
