//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//############################################################################
#include "ez_uail_gui.hpp"

//############################################################################
void ez_uail_point::to_ofstream_wave(ofstream& out,
                                     ez_wave_header& header,
                                     const bool& end_of_frame,
                                     const bool& unique_frame
                                    ) const
{
    short  xx, yy, zz, color;
    u_char rr, gg, bb;
    //------------------------------------------------------------------------
    xx =  x &  short_bit_mask[AIL->signal_bit_mask[0]];
    yy =  y &  short_bit_mask[AIL->signal_bit_mask[1]];
    zz =  z &  short_bit_mask[AIL->signal_bit_mask[5]];
    //------------------------------------------------------------------------
    rr = (is_blank()) ? (0) : (r & (short_bit_mask[AIL->signal_bit_mask[2]] >> 7));
    gg = (is_blank()) ? (0) : (g & (short_bit_mask[AIL->signal_bit_mask[3]] >> 7));
    bb = (is_blank()) ? (0) : (b & (short_bit_mask[AIL->signal_bit_mask[4]] >> 7));
    //------------------------------------------------------------------------
    out.put(    xx & 0x00ff      ); // x
    out.put((   xx & 0xff00) >> 8);
    //------------------------------------------------------------------------
    out.put(    yy & 0x00ff      ); // y
    out.put((   yy & 0xff00) >> 8);
    //------------------------------------------------------------------------
    color = (rr << 7) + end_of_frame;
    out.put( color & 0x00ff      ); // red
    out.put((color & 0xff00) >> 8);
    //------------------------------------------------------------------------
    color = (gg << 7) + unique_frame;
    out.put( color & 0x00ff      ); // green
    out.put((color & 0xff00) >> 8);
    //------------------------------------------------------------------------
    color = (bb << 7) + (bool)(k & EZ_UAIL_UNIQUE_POINT);
    out.put( color & 0x00ff      ); // blue
    out.put((color & 0xff00) >> 8);
    //------------------------------------------------------------------------
    switch(abs(header.signal_id[5]))
    {
        case EZ_SIGNAL_NO_SIGNAL:
                out.put(0x00);
                out.put(0x00);
                break;
        //--------------------------------------------------------------------
        case EZ_SIGNAL_Z_POSITION:
                out.put(    zz & 0x00ff      ); // z
                out.put((   zz & 0xff00) >> 8);
                break;
        //--------------------------------------------------------------------
        case EZ_SIGNAL_MONO_TTL:
                color = (is_lit()) ? (EZ_MAX_COLOR_SHORT) : (0);
                out.put( color & 0x00ff      );
                out.put((color & 0xff00) >> 8);
                break;
        //--------------------------------------------------------------------
        case EZ_SIGNAL_MONO_OR_ANALOG:
                color = ((rr | gg | bb) << 7);
                out.put( color & 0x00ff      );
                out.put((color & 0xff00) >> 8);
                break;
        //--------------------------------------------------------------------
        case EZ_SIGNAL_MONO_WEIGHTED_ANALOG:
                color = (((76 * rr) + (150 * gg) + (28 * bb)) / 254) << 7;
                out.put( color & 0x00ff      );
                out.put((color & 0xff00) >> 8);
                break;
        //--------------------------------------------------------------------
        case EZ_SIGNAL_MONO_AVG_ANALOG:
                color = ((rr + gg + bb) / 3) << 7;
                out.put( color & 0x00ff      );
                out.put((color & 0xff00) >> 8);
                break;
        //--------------------------------------------------------------------
    }
    header.num_samples++;
    return;
}

//############################################################################
void ez_uail_point::to_ofstream_wave_inverted(ofstream& out,
                                              ez_wave_header& header,
                                              const bool& end_of_frame,
                                              const bool& unique_frame
                                             ) const
{
    short  xx, yy, zz, color;
    u_char rr, gg, bb;
    //------------------------------------------------------------------------
    xx =  x &  short_bit_mask[AIL->signal_bit_mask[0]];
    yy =  y &  short_bit_mask[AIL->signal_bit_mask[1]];
    zz =  z &  short_bit_mask[AIL->signal_bit_mask[5]];
    //------------------------------------------------------------------------
    rr = (is_blank()) ? (0) : (r & (short_bit_mask[AIL->signal_bit_mask[2]] >> 7));
    gg = (is_blank()) ? (0) : (g & (short_bit_mask[AIL->signal_bit_mask[3]] >> 7));
    bb = (is_blank()) ? (0) : (b & (short_bit_mask[AIL->signal_bit_mask[4]] >> 7));
    //------------------------------------------------------------------------
    out.put(   -xx & 0x00ff      ); // x
    out.put((  -xx & 0xff00) >> 8);
    //------------------------------------------------------------------------
    out.put(   -yy & 0x00ff      ); // y
    out.put((  -yy & 0xff00) >> 8);
    //------------------------------------------------------------------------
    color = (rr << 7) + end_of_frame;
    color = -color;
    out.put( color & 0x00ff      ); // red
    out.put((color & 0xff00) >> 8);
    //--------------------------------------------------------------------
    color = (gg << 7) + unique_frame;
    color = -color;
    out.put( color & 0x00ff      ); // green
    out.put((color & 0xff00) >> 8);
    //--------------------------------------------------------------------
    color = (bb << 7) + (bool)(k & EZ_UAIL_UNIQUE_POINT);
    color = -color;
    out.put( color & 0x00ff      ); // blue
    out.put((color & 0xff00) >> 8);
    //--------------------------------------------------------------------
    switch(abs(header.signal_id[5]))
    {
        case EZ_SIGNAL_NO_SIGNAL:
                out.put(0x00);
                out.put(0x00);
                break;
        //--------------------------------------------------------------------
        case EZ_SIGNAL_Z_POSITION:
                out.put(   -zz & 0x00ff      ); // z
                out.put((  -zz & 0xff00) >> 8);
                break;
        //--------------------------------------------------------------------
        case EZ_SIGNAL_MONO_TTL:
                color = (is_lit()) ? (-EZ_MAX_COLOR_SHORT) : (0);
                out.put( color & 0x00ff      );
                out.put((color & 0xff00) >> 8);
                break;
        //--------------------------------------------------------------------
        case EZ_SIGNAL_MONO_OR_ANALOG:
                color = ((rr | gg | bb) << 7);
                color = -color;
                out.put( color & 0x00ff      );
                out.put((color & 0xff00) >> 8);
                break;
        //--------------------------------------------------------------------
        case EZ_SIGNAL_MONO_WEIGHTED_ANALOG:
                color = (((76 * rr) + (150 * gg) + (28 * bb)) / 254) << 7;
                color = -color;
                out.put( color & 0x00ff      );
                out.put((color & 0xff00) >> 8);
                break;
        //--------------------------------------------------------------------
        case EZ_SIGNAL_MONO_AVG_ANALOG:
                color = ((rr + gg + bb) / 3) << 7;
                color = -color;
                out.put( color & 0x00ff      );
                out.put((color & 0xff00) >> 8);
                break;
    //------------------------------------------------------------------------
    }
    header.num_samples++;
    return;
}

//############################################################################
ez_3D_short ez_uail_point::bit_masked_position() const
{
    ez_3D_short masked = *this;
    masked.x &= short_bit_mask[AIL->signal_bit_mask[0]];
    masked.y &= short_bit_mask[AIL->signal_bit_mask[1]];
    masked.z &= short_bit_mask[AIL->signal_bit_mask[5]];
    return masked;
}

//############################################################################
ez_color ez_uail_point::bit_masked_color() const
{
    ez_color masked = *this;
    masked.r &= (short_bit_mask[AIL->signal_bit_mask[2]] >> 7);
    masked.g &= (short_bit_mask[AIL->signal_bit_mask[3]] >> 7);
    masked.b &= (short_bit_mask[AIL->signal_bit_mask[4]] >> 7);
    return masked;
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
