//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//############################################################################
#ifndef __EZ_3D_POINT_DEFINITIONS__
#define __EZ_3D_POINT_DEFINITIONS__

//############################################################################
#include "ez_real_point.hpp"
#include "ez_wave.hpp"

//############################################################################
class ez_uail_point : public ez_3D_short, public ez_color
{
public:
    ez_uail_point( const short&   _x = 0,
                   const short&   _y = 0,
                   const short&   _z = 0,
                   const u_char&  _r = 255,
                   const u_char&  _g = 255,
                   const u_char&  _b = 255,
                   const u_char&  _k = EZ_UAIL_BLANKING_BIT | EZ_UAIL_UNIQUE_POINT,
                   const u_char&  _c = 55
                 )
                 : ez_3D_short (_x, _y, _z),
                   ez_color    (_r, _g, _b),
                   k           (_k        ),
                   c           (_c        )                     {}
    //------------------------------------------------------------------------
    ez_uail_point(const ez_uail_point& p)
                 : ez_3D_short (p.x, p.y, p.z),
                   ez_color    (p.r, p.g, p.b),
                   k           (p.k          ),
                   c           (p.c          )                  {}
    //------------------------------------------------------------------------
    ez_uail_point(const ez_real_point& r) // cast to ez_uail_point
                 : ez_3D_short ((short)round(r.x), (short)round(r.y), (short)round(r.z)),
                   ez_color    (r.r, r.g, r.b),
                   k           (r.k          ),
                   c           (r.c          )                  {}
    //------------------------------------------------------------------------
    ez_uail_point( const ez_3D_short& s,
                   const ez_color&  rgb,
                   const u_char&    _k = EZ_UAIL_BLANKING_BIT | EZ_UAIL_UNIQUE_POINT,
                   const u_char&    _c = 0
                 )
                 : ez_3D_short (s  ),
                   ez_color    (rgb),
                   k           (_k ),
                   c           (_c )                            {}
    //------------------------------------------------------------------------
    ez_uail_point( const ez_3D_short& s,
                   const u_char&     _c
                 )
                 : ez_3D_short (s  ),
                   ez_color    (255, 255, 255),
                   k           (EZ_UAIL_BLANKING_BIT),
                   c           (_c )                            {}
    //------------------------------------------------------------------------
    ez_uail_point( const ez_3D_short& s)
                 : ez_3D_short (s                ),
                   ez_color    (255, 255, 255    ),
                   k           (EZ_UAIL_BLANKING_BIT | EZ_UAIL_UNIQUE_POINT),
                   c           (0                )              {}
    //------------------------------------------------------------------------
    ez_uail_point( const ez_color& rgb)
                 : ez_3D_short (0, 0, 0          ),
                   ez_color    (rgb              ),
                   k           (EZ_UAIL_BLANKING_BIT | EZ_UAIL_UNIQUE_POINT),
                   c           (0                )              {}
    //------------------------------------------------------------------------
    ez_uail_point& operator = (const ez_uail_point& p)
                    {
                        x = p.x;
                        y = p.y;
                        z = p.z;
                        r = p.r;
                        g = p.g;
                        b = p.b;
                        k = p.k;
                        c = p.c;
                        return *this;
                    }
    //------------------------------------------------------------------------
    ez_uail_point& operator = (const ez_3D_short& s) // only assign the coordinates
                    {
                        x = s.x;
                        y = s.y;
                        z = s.z;
                        return *this;
                    }
    //------------------------------------------------------------------------
    ez_uail_point  operator + (const ez_3D_short& s)
                    {
                        ez_uail_point sum(*this);
                        sum.x += s.x;
                        sum.y += s.y;
                        sum.z += s.z;
                        return sum;
                    }
    //------------------------------------------------------------------------
    ez_uail_point& operator += (const ez_3D_short& s)
                    {
                        x += s.x;
                        y += s.y;
                        z += s.z;
                        return *this;
                    }
    //------------------------------------------------------------------------
    ez_uail_point  operator - (const ez_3D_short& s)
                    {
                        ez_uail_point diff(*this);
                        diff.x -= s.x;
                        diff.y -= s.y;
                        diff.z -= s.z;
                        return diff;
                    }
    //------------------------------------------------------------------------
    ez_uail_point& operator -= (const ez_3D_short& s)
                    {
                        x -= s.x;
                        y -= s.y;
                        z -= s.z;
                        return *this;
                    }
    //------------------------------------------------------------------------
    ez_uail_point& operator *= (const ez_3D_short& s)
                    {
                        x = (short)(x * s.x);
                        y = (short)(y * s.y);
                        z = (short)(z * s.z);
                        return *this;
                    }
    //------------------------------------------------------------------------
    ez_uail_point& operator *= (const ez_3D_double& f)
                    {
                        x = (short)(x * f.x);
                        y = (short)(y * f.y);
                        z = (short)(z * f.z);
                        return *this;
                    }
    //------------------------------------------------------------------------
    bool operator == (const ez_uail_point& p) const
            {   
                if(    (x != p.x)
                    || (y != p.y)
                    || (z != p.z)
                    || (r != p.r)
                    || (g != p.g)
                    || (b != p.b)
                    || ((k & EZ_UAIL_BLANKING_BIT) != (p.k & EZ_UAIL_BLANKING_BIT))
                    || (c != p.c)
                  )
                    return false;
                return true;
            }
    //------------------------------------------------------------------------
    bool operator != (const ez_uail_point& p) const
            {                               
                if(    (x == p.x)
                    && (y == p.y)
                    && (z == p.z)
                    && (r == p.r)
                    && (g == p.g)
                    && (b == p.b)
                    && ((k & EZ_UAIL_BLANKING_BIT) == (p.k & EZ_UAIL_BLANKING_BIT))
                    && (c == p.c)
                  )
                    return false;
                return true;
            }
    //------------------------------------------------------------------------
    ez_uail_point  operator - ()
                    {
                        ez_uail_point negative;
                        negative.x = -x;
                        negative.y = -y;
                        negative.z = -z;
                        return negative;
                    }
    //------------------------------------------------------------------------
    bool is_equal_2D(const ez_uail_point& p) const
            {   
                if(    (x != p.x)
                    || (y != p.y)
                    || (r != p.r)
                    || (g != p.g)
                    || (b != p.b)
                    || ((k & EZ_UAIL_BLANKING_BIT) != (p.k & EZ_UAIL_BLANKING_BIT))
                    || (c != p.c)
                  )
                    return false;
                return true;
            }
    //------------------------------------------------------------------------
    int from_ifstream_ilda (ifstream& in, char format)
            {
                if(!in.eof())
                {
                    char p, q;
                    //--------------------------------------------------------
                    in.get(p);
                    in.get(q);
                    x = ((p & 0x00ff) << 8 | (q & 0x00ff));
                    //--------------------------------------------------------
                    in.get(p);
                    in.get(q);
                    y = ((p & 0x00ff) << 8 | (q & 0x00ff));
                    //--------------------------------------------------------
                    if(    format == EZ_UAIL_3D_VECTOR
                        || format == EZ_UAIL_3D_COLOR_VECTOR
                      )
                    {
                        in.get(p);
                        in.get(q);
                        z = ((p & 0x00ff) << 8 | (q & 0x00ff));
                    }
                    //--------------------------------------------------------
                    k = (u_char)in.get();
                    if(    format == EZ_UAIL_3D_VECTOR
                        || format == EZ_UAIL_2D_VECTOR
                      )
                    {
                        c = (u_char)in.get();
                    }
                    else if(    format == EZ_UAIL_3D_COLOR_VECTOR
                             || format == EZ_UAIL_2D_COLOR_VECTOR
                           )
                    {
                        b = (u_char)in.get();
                        g = (u_char)in.get();
                        r = (u_char)in.get();
                    }
                    //--------------------------------------------------------
                    set_unique();
                }
                return in.bad();
            }
    //------------------------------------------------------------------------
    void to_ofstream_ilda(ofstream& out, char format) const
            {
                switch(format)
                {                              
                    case EZ_UAIL_3D_VECTOR:
                             out.put((char)((x & 0xff00) >> 8));
                             out.put((char) (x & 0x00ff)      );
                             out.put((char)((y & 0xff00) >> 8));
                             out.put((char) (y & 0x00ff)      );
                             out.put((char)((z & 0xff00) >> 8));
                             out.put((char) (z & 0x00ff)      );
                             out.put((char)k);
                             out.put((char)c);                     
                             break;
                    //--------------------------------------------------------
                    case EZ_UAIL_2D_VECTOR:
                             out.put((char)((x & 0xff00) >> 8));
                             out.put((char) (x & 0x00ff)      );
                             out.put((char)((y & 0xff00) >> 8));
                             out.put((char) (y & 0x00ff)      );
                             out.put((char)k);
                             out.put((char)c);
                             break;
                    //--------------------------------------------------------
                    case EZ_UAIL_3D_COLOR_VECTOR:
                             out.put((char)((x & 0xff00) >> 8));
                             out.put((char) (x & 0x00ff)      );
                             out.put((char)((y & 0xff00) >> 8));
                             out.put((char) (y & 0x00ff)      );
                             out.put((char)((z & 0xff00) >> 8));
                             out.put((char) (z & 0x00ff)      );
                             out.put((char)k);
                             out.put((char)b);
                             out.put((char)g);
                             out.put((char)r);
                             break;
                    //--------------------------------------------------------
                    case EZ_UAIL_2D_COLOR_VECTOR:
                             out.put((char)((x & 0xff00) >> 8));
                             out.put((char) (x & 0x00ff)      );
                             out.put((char)((y & 0xff00) >> 8));
                             out.put((char) (y & 0x00ff)      );
                             out.put((char)k);
                             out.put((char)b);
                             out.put((char)g);
                             out.put((char)r);
                             break;
                    //--------------------------------------------------------
                }
                return;
            }
    //------------------------------------------------------------------------
    void  to_ofstream_wave(ofstream& out,
                           ez_wave_header& header,
                           const bool& end_of_frame = false,
                           const bool& unique_frame = false
                          ) const;
    //------------------------------------------------------------------------
    void  to_ofstream_wave_inverted(ofstream& out,
                                    ez_wave_header& header,
                                    const bool& end_of_frame = false,
                                    const bool& unique_frame = false
                                   ) const;
    //------------------------------------------------------------------------
    void  select       ()       {  k |=  EZ_UAIL_SELECTED_POINT;      return; }
    void  deselect     ()       {  k &= ~EZ_UAIL_SELECTED_POINT;      return; }
    void  toggle_select()       {  is_selected() ? deselect() : select();     }
    bool  is_selected  () const {  return (bool)(k & EZ_UAIL_SELECTED_POINT); }
    //------------------------------------------------------------------------
    ez_uail_point& blank   ()   {  k |=  EZ_UAIL_BLANKING_BIT; return *this;  }
    ez_uail_point& unblank ()   {  k &= ~EZ_UAIL_BLANKING_BIT; return *this;  }
    //------------------------------------------------------------------------
    bool  is_blank     () const {  return (bool)(k & EZ_UAIL_BLANKING_BIT);   }
    bool  is_lit       () const {  return !is_blank();                        }
    //------------------------------------------------------------------------
    void  set_unique   ()       {  k |=  EZ_UAIL_UNIQUE_POINT;      return;   }
    void  clear_unique ()       {  k &= ~EZ_UAIL_UNIQUE_POINT;      return;   }
    bool  is_unique    () const {  return (bool)(k & EZ_UAIL_UNIQUE_POINT);   }
    //------------------------------------------------------------------------
    bool  is_black () const
                    {
                        return (    is_lit()
                                 && r == 0
                                 && g == 0
                                 && b == 0
                               );
                    }
    //------------------------------------------------------------------------
    bool  is_color () const
            {
                return (    is_lit()
                         && (r + g + b)
                       );
            }
    //------------------------------------------------------------------------
    void  clear()
            {
                x = 0;
                y = 0;
                z = 0;
                r = 0;
                g = 0;
                b = 0;
                k = EZ_UAIL_BLANKING_BIT | EZ_UAIL_UNIQUE_POINT;
                c = 0;
            }
    //------------------------------------------------------------------------
    ez_3D_short position() const    {    return (ez_3D_short)(*this);    }
    //------------------------------------------------------------------------
    int color() const
        {
            return   is_lit()
                   ? (int)(   r << 16
                            | g <<  8
                            | b
                          )
                   : -1; // is blank
        }
    //------------------------------------------------------------------------
    ez_color as_ez_color() const
        {
            return (ez_color)(*this);
        }
    //------------------------------------------------------------------------
    ez_3D_short  bit_masked_position() const;
    ez_color     bit_masked_color()    const;
    //------------------------------------------------------------------------
    u_char  k, c;
};

//############################################################################
#endif

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
